"""
Driver for running model on design of experiments cases using Salib sampling methods
"""
from __future__ import print_function
from six import itervalues, iteritems, reraise
from six.moves import range

import numpy as np

from openmdao.api import DOEDriver, ListGenerator
from openmdao.core.driver import Driver, RecordingDebugging
from openmdao.drivers.doe_generators import DOEGenerator
from openmdao.utils.general_utils import warn_deprecation

SALIB_NOT_INSTALLED = False
try:
    from SALib.sample import morris as ms
except ImportError:
    SALIB_NOT_INSTALLED = True

class SalibMorrisDOEGenerator(DOEGenerator):

    def __init__(self, n_trajs=2, n_levels=4):
        super(SalibMorrisDOEGenerator, self).__init__()

        if SALIB_NOT_INSTALLED:
            raise RuntimeError('SALib library is not installed. \
                                cf. https://salib.readthedocs.io/en/latest/getting-started.html')

        # number of trajectories to apply morris method
        self.n_trajs = n_trajs
        # number of grid levels
        self.n_levels = n_levels

    def __call__(self, design_vars, model=None):
        bounds=[]
        names=[]
        for name, meta in iteritems(design_vars):
            size = meta['size']
            meta_low = meta['lower']
            meta_high = meta['upper']
            for j in range(size):
                if isinstance(meta_low, np.ndarray):
                    p_low = meta_low[j]
                else:
                    p_low = meta_low

                if isinstance(meta_high, np.ndarray):
                    p_high = meta_high[j]
                else:
                    p_high = meta_high
                    
                display_name = name.split('.')[-1]
                if size>1:
                    display_name += str(j)
                names.append(display_name)
                bounds.append((p_low, p_high))

        self._pb = {'num_vars': len(names), 
                    'names': names, 
                    'bounds': bounds, 'groups': None}
        cases = ms.sample(self._pb, self.n_trajs, self.n_levels)
        sample = []
        for i in range(cases.shape[0]):
            j=0
            for name, meta in iteritems(design_vars):
                size = meta['size']
                sample.append((name, cases[i, j:j + size]))
                j += size
            yield sample

class SalibMorrisDOEDriver(DOEDriver):
    """
    Baseclass for SALib design-of-experiments Drivers
    """

    def __init__(self, **kwargs):
        super(SalibMorrisDOEDriver, self).__init__()

        if SALIB_NOT_INSTALLED:
            raise RuntimeError('SALib library is not installed. \
                                cf. https://salib.readthedocs.io/en/latest/getting-started.html')

        self.options.declare('n_trajs', types=int, default=2,
                             desc='number of trajectories to apply morris method')
        self.options.declare('n_levels', types=int, default=4,
                             desc='number of grid levels')
        self.options.update(kwargs)
        n_trajs = self.options['n_trajs']
        n_levels = self.options['n_levels']
        self.options['generator'] = SalibMorrisDOEGenerator(n_trajs, n_levels)

class SalibDoeDriver(SalibMorrisDOEDriver):
    """
    Deprecated. Use SalibMorrisDOEDriver.
    """
    def __init__(self, **kwargs):
        super(SalibDoeDriver, self).__init__(**kwargs)
        warn_deprecation("'SalibDoeDriver' is deprecated; "
                         "use 'SalibMorrisDOEDriver' instead.")

# class SalibSobolDOEGenerator(DOEGenerator):

#     def __init__(self, ):
#         super(SalibSobolDOEGenerator, self).__init__()

#         if SALIB_NOT_INSTALLED:
#             raise RuntimeError('SALib library is not installed. \
#                                 cf. https://salib.readthedocs.io/en/latest/getting-started.html')

#         # number of trajectories to apply morris method
#         self.n_trajs = n_trajs
#         # number of grid levels
#         self.n_levels = n_levels
#         # grid jump size
#         self.grid_step_size = grid_step_size        

#     def __call__(self, design_vars, model=None):
#         bounds=[]
#         names=[]
#         for name, meta in iteritems(design_vars):
#             size = meta['size']
#             meta_low = meta['lower']
#             meta_high = meta['upper']
#             for j in range(size):
#                 if isinstance(meta_low, np.ndarray):
#                     p_low = meta_low[j]
#                 else:
#                     p_low = meta_low

#                 if isinstance(meta_high, np.ndarray):
#                     p_high = meta_high[j]
#                 else:
#                     p_high = meta_high
                    
#                 display_name = name.split('.')[-1]
#                 if size>1:
#                     display_name += str(j)
#                 names.append(display_name)
#                 bounds.append((p_low, p_high))

#         self._pb = {'num_vars': len(names), 
#                     'names': names, 
#                     'bounds': bounds, 'groups': None}
#         cases = ms.sample(self._pb, self.n_trajs, self.n_levels, self.grid_step_size)
#         sample = []
#         for i in range(cases.shape[0]):
#             j=0
#             for name, meta in iteritems(design_vars):
#                 size = meta['size']
#                 sample.append((name, cases[i, j:j + size]))
#                 j += size
#             yield sample



from __future__ import print_function
import numpy as np
from six import iteritems

from openmdao.core.driver import Driver, RecordingDebugging
from openmdao.core.analysis_error import AnalysisError
from openmdao.recorders.recording_iteration_stack import Recording

ONERASEGO_NOT_INSTALLED = False
try:
    from segomoe.sego_defs import get_sego_options, ExitStatus
    from segomoe.constraint import Constraint
    from segomoe.sego import Sego
except ImportError:
    ONERASEGO_NOT_INSTALLED = True

def to_list(l, size):
    if not (isinstance(l, np.ndarray) or isinstance(l, list)):
        return [l] * size
    diff_len = len(l) - size
    if diff_len > 0:
        return l[0:size]
    elif diff_len < 0:
        return [l[0]] * size
    else:
        return l

class OneraSegoDriver(Driver):
    """
    OpenMDAO driver for ONERA SEGOMOE optimizer 
    """

    def __init__(self, **kwargs):
        """
        Initialize the driver with the following option.
        """
        super(OneraSegoDriver, self).__init__(**kwargs)

        if ONERASEGO_NOT_INSTALLED:
            raise RuntimeError('Onera SEGOMOE library is not installed.')

        # What we support
        self.supports['inequality_constraints'] = True
        self.supports['equality_constraints'] = True
        self.supports['two_sided_constraints'] = True
        self.supports['linear_constraints'] = True

        # What we don't support
        self.supports['multiple_objectives'] = False
        self.supports['active_set'] = False
        self.supports['integer_design_vars'] = False
        self.supports['simultaneous_derivatives'] = False
        self.supports['total_jac_sparsity'] = False
        self.supports['gradients'] = False
        
        self.opt_settings = {}

    def _declare_options(self):
        self.options.declare('optimizer', default='SEGOMOE', values=['SEGOMOE'],
                             desc='Name of optimizers to use')
        
    def _setup_driver(self, problem):
        super(OneraSegoDriver, self)._setup_driver(problem)
        
        self.comm = None

    def run(self, path_hs='', eq_tol={}, ieq_tol={}):
        """
        Optimize the problem using SEGOMOE.

        Parameters
        ----------
        problem : Problem object
            An optimization problem of OpenMDAO framework
        path_hs: string, optional
            path to a directory storing hot_start data
        eq_tol: dict
            Dictionary to define specific tolerance for eq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        ieq_tol: dict
            Dictionary to define specific tolerance for ieq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        """

        model = self._problem.model
        path_hs=''
        self.eq_tol=eq_tol
        self.ieq_tol=ieq_tol
        self.iter_count = 0
        self.name = optimizer = "onera_optimizer_segomoe"
        self._sego_vars = []
        self._sego_cons = []
        self._map_con = {}
        self._n_mapped_con = 0
        
        # Initial Run
        with RecordingDebugging(self.options['optimizer'], self.iter_count, self) as rec:
            # Initial Run
            model._solve_nonlinear()
            rec.abs = 0.0
            rec.rel = 0.0
        self.iter_count += 1

        # Format design variables to suit segomoe implementation
        self._initialize_vars()
        
        # Format constraints to suit segomoe implementation
        self._initialize_cons()
        
        # Format option dictionary to suit SEGO implementation
        if 'n_iter' not in self.opt_settings:
            self.opt_settings.update({'n_iter': 100}) # not part of segomoe options per se but an option of run_optim
        optim_settings = {}
        for opt, opt_dict in iteritems(get_sego_options()):
            optim_settings[opt] = opt_dict['default']
        optim_settings.update(self.opt_settings)
        del optim_settings['n_iter']  
        
        # In OpenMDAO context, obj and constraints are always evaluated together
        optim_settings['grouped_eval']=True

        # default model
        mod_obj = {'corr': 'squared_exponential',
                   'regr': 'constant',
                   'normalize': True}
        
        dim=0
        for name, meta in iteritems(self._designvars):
            dim += meta['size']
        print("Designvars dimension: ", dim)
        if dim > 10:
            n_components = 3
            mod_obj['n_components'] = n_components
            mod_obj['type'] = 'KrigKPLS' if dim > 20 else 'KrigKPLSK'
        else:
            n_components = dim
            mod_obj['type'] = 'Krig'
        mod_obj['theta0'] = [1.0] * n_components
        mod_obj['thetaL'] = [0.1] * n_components
        mod_obj['thetaU'] = [10.0] * n_components

        optim_settings['model_type']= {'obj': mod_obj, 'con': mod_obj}

        # Instanciate a SEGO optimizer
        sego = Sego(self._objfunc, self._sego_vars, const=self._sego_cons, \
                    optim_settings=optim_settings, path_hs=path_hs, comm=self.comm)

        # Run the optim
        exit_flag, x_best, obj_best, dt_opt = sego.run_optim(n_iter=self.opt_settings['n_iter'])

        # Set optimal parameters
        i = 0
        for name, meta in iteritems(self._designvars):
            size = meta['size']
            self.set_design_var(name, x_best[i: i + size])
            i += size

        with RecordingDebugging(self.options['optimizer'], self.iter_count, self) as rec:
            model._solve_nonlinear()
            rec.abs = 0.0
            rec.rel = 0.0
        self.iter_count += 1

        self.exit_flag = (exit_flag == ExitStatus.valid_sol) or \
                         (exit_flag == ExitStatus.solution_reached)
                         
        return self.exit_flag

    def _initialize_vars(self):
        vars = []
        param_meta = self._designvars
        param_vals = self.get_design_var_values()
        for name, meta in iteritems(param_meta):
            if meta['size'] > 1:
                lower = meta['lower'] if np.isscalar(meta['lower']) else meta['lower'][i]
                upper = meta['upper'] if np.isscalar(meta['lower']) else meta['upper'][i]
                vars += [{'name': name + '_' + str(i),
                         'lb': lower,
                         'ub': upper} for i in range(meta['size'])]
            else:
                vars += [{'name': name, 'lb': meta['lower'], 'ub': meta['upper']}]
        self._sego_vars = vars

    def _initialize_cons(self):
        """
        Format OpenMDAO constraints to suit SEGOMOE implementation

        Parameters
        ----------
        eq_tol: dict
            Dictionary to define specific tolerance for eq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        ieq_tol: dict
            Dictionary to define specific tolerance for ieq constraints
            {'[groupName]': [tol]} Default tol = 1e-5
        """

        # for short
        add_con = self._sego_cons.append

        # Start at 1 -> obj first
        eval_index = 1

        con_meta = self._cons
        eq_cons = {name: con for name, con in iteritems(con_meta) if con['equals']}
        ieq_cons = {name: con for name, con in iteritems(con_meta) if not con['equals']}
        
        # Equality constraints
        for name in eq_cons.keys():
            meta = con_meta[name]
            size = meta['size']
            equals = to_list(meta['equals'], size)
            tol = to_list(eq_tol[name] if name in self.eq_tol else 1e-4, size)

            for k in xrange(size):
                add_con(Constraint('=', equals[k], name=(name + '_' + str(k)), \
                                   tol=tol[k]))
            # Map eq constraints - range to allow double-sided constraint support
            self._map_con[name] = range(eval_index, eval_index + size)
            eval_index += size
            self._n_mapped_con += size

        # Inequality constraints
        for name in ieq_cons.keys():
            meta = con_meta[name]
            size = meta['size']
            # Bounds - double sided is supported
            lower = to_list(meta['lower'], size)
            upper = to_list(meta['upper'], size)
            tol = to_list(ieq_tol[name] if name in self.ieq_tol else 1e-4, size)
            # Index list is updated for each real constraint to handle double sided constraints
            index_list = []

            for k in range(size):
                if lower[k] is None or lower[k] < -1e29:
                    # g(x) < upper
                    add_con(Constraint('<', upper[k], name=(name + '_' + str(k)), \
                                       tol=tol[k]))
                    index_list.append(eval_index)
                    eval_index += 1
                    self._n_mapped_con += 1
                elif upper[k] is None or upper[k] > 1e29:
                    # g(x) > lower
                    add_con(Constraint('>', lower[k], name=(name + '_' + str(k)), \
                                       tol=tol[k]))
                    index_list.append(eval_index)
                    eval_index += 1
                    self._n_mapped_con += 1
                else:
                    # lower < g(x) < upper
                    # Split is currently required for sego
                    add_con(Constraint('<', upper[k], name=(name + '_' + str(k) + '<'), \
                                       tol=tol[k]))
                    add_con(Constraint('>', lower[k], name=(name + '_' + str(k) + '>'), \
                                       tol=tol[k]))
                    index_list.append([eval_index, eval_index + 1])
                    eval_index += 2
                    self._n_mapped_con += 2
            self._map_con[name] = index_list
            
                
    def _objfunc(self, point):
        """
        Function that evaluates and returns the objective function and the
        constraints. This function is called by SEGOMOE

        Parameters
        ----------
        point : numpy.ndarray
            point to evaluate

        Returns
        -------
        func_dict : dict
            Dictionary of all functional variables evaluated at design point.
        fail : int
            0 for successful function evaluation
            1 for unsuccessful function evaluation
        """
        fail = False
        res = np.zeros(1 + self._n_mapped_con)
        model = self._problem.model

        try:
            # Pass in new parameters
            i = 0

            for name, meta in iteritems(self._designvars):
                size = meta['size']
                self.set_design_var(name, point[i: i + size])
                i += size

            # Execute the model
            with RecordingDebugging(self.options['optimizer'], self.iter_count, self) as rec:
                self.iter_count += 1
                try:
                    model._solve_nonlinear()

                    # Let the optimizer try to handle the error
                except AnalysisError:
                    model._clear_iprint()
                    fail = True

            # Get the objective function evaluation - single obj support
            for name, obj in iteritems(self.get_objective_values()):
                res[0] = obj

            # Get the constraint evaluations:
            for name, con_res in iteritems(self.get_constraint_values()):
                # Make sure con_res is array_like
                con_res = to_list(con_res, len(self._map_con[name]))
                # Perform mapping
                for i, con_index in enumerate(self._map_con[name]):
                    if isinstance(con_index, list):
                        # Double sided inequality constraint -> duplicate response
                        for k in con_index:
                            res[k] = con_res[i]
                    else:
                        res[con_index] = con_res[i]

        except Exception as msg:
            tb = traceback.format_exc()

            # Exceptions seem to be swallowed by the C code, so this
            # should give the user more info than the dreaded "segfault"
            print("Exception: %s" % str(msg))
            print(70 * "=", tb, 70 * "=")
            fail = True

        return res, fail



'''
Author: Remi Lafage <remi.lafage@onera.fr>

This package is distributed under Apache 2 license.
'''

from setuptools import setup

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

CLASSIFIERS = """
Development Status :: 4 - Beta
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: Apache Software License
Programming Language :: Python
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3.6
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: Microsoft :: Windows
Operating System :: Unix
Operating System :: MacOS
"""

metadata = dict(
    name='openmdao_extensions',
    version='0.2.0',
    description='Additional solvers and drivers for OpenMDAO framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rémi Lafage',
    author_email='remi.lafage@onera.fr',
    license='Apache License, Version 2.0',
    classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
    packages=[
        'openmdao_extensions'
    ],
    install_requires=[
        'openmdao'
    ],
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*',
    zip_safe=True,
    url = 'https://github.com/OneraHub/opendmao_extensions',
    download_url = 'https://github.com/OneraHub/openmdao_extensions/releases',
)

setup(**metadata)



"""
Driver for running model on design of experiments cases using SMT sampling methods
"""
from __future__ import print_function
from six import itervalues, iteritems, reraise
import numpy as np

from openmdao.api import DOEDriver, ListGenerator
from openmdao.core.driver import Driver, RecordingDebugging
from openmdao.drivers.doe_generators import DOEGenerator
from openmdao.utils.general_utils import warn_deprecation

SMT_NOT_INSTALLED = False
try:
    from smt.sampling_methods import FullFactorial, LHS, Random
except:
    SMT_NOT_INSTALLED = False

_sampling_methods = {'FullFactorial': FullFactorial, 'LHS': LHS, 'Random': Random}
class SmtDOEGenerator(DOEGenerator):

    def __init__(self, sampling_method_name, n_cases, **kwargs):
        super(SmtDOEGenerator, self).__init__()

        if SMT_NOT_INSTALLED:
            raise RuntimeError('SMT library is not installed. cf. https://https://smt.readthedocs.io/en/latest')

        # number of trajectories to apply morris method
        self.sampling_method_name = sampling_method_name
        # number of grid levels
        self.n_cases = n_cases
        # options
        self.sampling_method_opts = kwargs

    def __call__(self, design_vars, model=None):
        xlimits=[]
        for name, meta in iteritems(design_vars):
            size = meta['size']
            meta_low = meta['lower']
            meta_high = meta['upper']
            for j in range(size):
                if isinstance(meta_low, np.ndarray):
                    p_low = meta_low[j]
                else:
                    p_low = meta_low

                if isinstance(meta_high, np.ndarray):
                    p_high = meta_high[j]
                else:
                    p_high = meta_high

                xlimits.append((p_low, p_high))

        sampling = _sampling_methods[self.sampling_method_name](xlimits=np.array(xlimits), **(self.sampling_method_opts))
        cases = sampling(self.n_cases)
        sample = []
        for i in range(cases.shape[0]):
            j=0
            for name, meta in iteritems(design_vars):
                size = meta['size']
                sample.append((name, cases[i, j:j + size]))
                j += size
            yield sample

class SmtDOEDriver(DOEDriver):
    """
    Baseclass for SMT design-of-experiments Drivers 
    """
    def __init__(self, **kwargs):
        super(SmtDOEDriver, self).__init__()

        if SMT_NOT_INSTALLED:
            raise RuntimeError('SMT library is not installed. cf. https://https://smt.readthedocs.io/en/latest')

        self.options.declare('sampling_method_name', types=str, default='LHS',
                             desc='either LHS, FullFactorial or Ramdom')
        self.options.declare('n_cases', types=int, default=2,
                             desc='number of sample to generate')
        self.options.declare('sampling_method_options', types=dict, default={},
                             desc='options for given SMT sampling method')

        self.options.update(kwargs)
        name = self.options['sampling_method_name']
        n_cases = self.options['n_cases']
        opts = self.options['sampling_method_options']
        self.options['generator'] = SmtDOEGenerator(sampling_method_name=name, n_cases=n_cases, **opts)

class SmtDoeDriver(SmtDOEDriver):
    """
    Deprecated. Use SmtDOEDriver  
    """
    def __init__(self, **kwargs):
        super(SmtDoeDriver, self).__init__(**kwargs)
        warn_deprecation("'SmtDoeDriver' is deprecated; "
                         "use 'SmtDOEDriver' instead.")




